use std::collections::HashMap;

use crate::{
    ast2::{
        intern::{self, AMD64_POINTER_TYPE_INFO},
        Index,
    },
    triples::{self, Inst, IR},
};

#[derive(Debug, thiserror::Error)]
enum Error {
    #[error("This visitor does not implement visiting this tag.")]
    Unimplemented,
}

use super::{
    intern::InternPoolWrapper as InternPool,
    visitor::{AstExt, AstVisitorTrait},
    Ast, Data, PlaceOrValue, Tag, TypeCache,
};

struct IrBuilder {
    functions: Vec<Function>,
}

struct Function {
    ir: IR,
}

struct IrFunctionBuilder<'a> {
    ir: &'a mut IR,
    type_cache: &'a mut TypeCache,
    ref_lookup: &'a mut HashMap<Index, u32>,
    ip: &'a InternPool,
    function: Function,
}

impl<'b, 'a> AstVisitorTrait<&'a mut super::Ast> for IrFunctionBuilder<'b> {
    type Error = Error;
    type Value = Option<u32>;
    const UNIMPL: Self::Error = Error::Unimplemented;

    fn visit_block_impl(
        &mut self,
        ast: &'a mut super::Ast,
        _idx: Index,
        statements: Vec<Index>,
        expr: Option<Index>,
    ) -> Result<Self::Value, Self::Error> {
        statements
            .into_iter()
            .map(|i| self.visit_any(ast, i))
            .try_fold((), |_, e| e.map(crate::unit))?;

        expr.map(|expr| self.visit_any(ast, expr))
            .transpose()
            .map(Option::flatten)
    }

    fn visit_place_to_value_conversion_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        expr: Index,
    ) -> Result<Self::Value, Self::Error> {
        // idx's (this) type is the pointee type
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, idx);

        let expr = self.visit_any(ast, expr)?.unwrap();

        let ir = self.ir.push(Inst::Load(ty), Some(triples::Data::lhs(expr)));
        Ok(Some(ir))
    }

    fn visit_value_to_place_conversion_impl(
        &mut self,
        ast: &'a mut super::Ast,
        _idx: Index,
        expr: Index,
    ) -> Result<Self::Value, Self::Error> {
        // expr's type is the pointee type
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, expr);

        let expr = self.visit_any(ast, expr)?.unwrap();

        let info = self.ip.size_of_type(ty, AMD64_POINTER_TYPE_INFO);

        let alloca = self.ir.push(
            Inst::Alloca,
            Some(triples::Data::new(info.size(), info.align())),
        );

        _ = self
            .ir
            .push(Inst::Store(ty), Some(triples::Data::new(expr, alloca)));

        Ok(Some(alloca))
    }

    fn visit_decl_ref_impl(
        &mut self,
        _ast: &'a mut super::Ast,
        _idx: Index,
        decl: Index,
    ) -> Result<Self::Value, Self::Error> {
        let alloca = self.ref_lookup.get(&decl).cloned().expect("declref");

        Ok(Some(alloca))
    }

    fn visit_var_decl_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        name: intern::Index,
        ty: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_var_decl_common(ast, idx)
    }

    fn visit_mut_var_decl_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        name: intern::Index,
        ty: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_var_decl_common(ast, idx)
    }

    fn visit_var_decl_assignment_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        name: intern::Index,
        expr: Index,
        ty: Option<Index>,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_var_decl_common(ast, idx)
    }

    fn visit_mut_var_decl_assignment_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        name: intern::Index,
        expr: Index,
        ty: Option<Index>,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_var_decl_common(ast, idx)
    }

    fn visit_return_stmt_impl(
        &mut self,
        _ast: &'a mut super::Ast,
        _idx: Index,
    ) -> Result<Self::Value, Self::Error> {
        let ir = self.ir.push(Inst::Return, None);
        Ok(Some(ir))
    }
    fn visit_return_expr_stmt_impl(
        &mut self,
        ast: &'a mut super::Ast,
        _idx: Index,
        expr: Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, expr);

        let expr = self.visit_any(ast, expr)?.unwrap();

        let ir = self
            .ir
            .push(Inst::ReturnValue(ty), Some(triples::Data::lhs(expr)));
        Ok(Some(ir))
    }

    fn visit_constant_impl(
        &mut self,
        ast: &'a mut super::Ast,
        _idx: Index,
        ty: Index,
        value: intern::Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, ty);
        let ir = self.ir.push(Inst::Constant, Some((ty, value).into()));

        Ok(Some(ir))
    }

    fn visit_subscript_expr_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        index: Index,
    ) -> Result<Self::Value, Self::Error> {
        // pointer type
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, idx);

        let lhs = self.visit_any(ast, lhs)?.unwrap();
        let index = self.visit_any(ast, index)?.unwrap();

        // TODO: make getelementptr take a variable instead of a constant index.
        let ir = self.ir.push(
            Inst::GetElementPtr(ty),
            Some(triples::Data::new(lhs, index)),
        );

        Ok(Some(ir))
    }

    fn visit_call_expr_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        func: Index,
        argument_list: Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, idx);

        let len = self.visit_argument_list(ast, argument_list)?.unwrap();
        _ = self.ir.push(Inst::InlineType(ty), None);
        let end = self.ir.nodes.len() as u32;

        let func = self.visit_any(ast, func)?.unwrap();

        let result = self
            .ir
            .push(Inst::Call(func), Some(triples::Data::new(end - len, end)));

        Ok(Some(result))
    }

    fn visit_argument_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        expr: Index,
        name: Option<intern::Index>,
    ) -> Result<Self::Value, Self::Error> {
        let arg = self.visit_any(ast, expr)?.unwrap();

        // let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, expr);
        // self.ir.push(Inst::Argument(ty), Some(triples::Data::lhs(arg)));

        Ok(Some(arg))
    }

    fn visit_argument_list_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        arguments: Vec<Index>,
    ) -> Result<Self::Value, Self::Error> {
        let args = arguments
            .into_iter()
            .map(|i| {
                let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, i);
                let arg = self.visit_argument(ast, i)?.unwrap();
                Ok((ty, arg))
            })
            .collect::<Result<Vec<_>, _>>()?;

        let count = args.len();
        for (ty, arg) in args {
            _ = self
                .ir
                .push(Inst::Argument(ty), Some(triples::Data::lhs(arg)));
        }
        // .try_fold((), |_, e| e.map(crate::unit))?;

        Ok(Some(count as u32))
    }

    fn visit_function_proto_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        name: intern::Index,
        return_type: Index,
        parameter_list: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_parameter_list(ast, parameter_list)?;
        // push parameters to ir
        Ok(None)
    }

    fn visit_parameter_list_impl(
        &mut self,
        ast: &'a mut super::Ast,
        _idx: Index,
        params: Vec<Index>,
    ) -> Result<Self::Value, Self::Error> {
        params
            .into_iter()
            .map(|i| self.visit_parameter(ast, i))
            .try_fold((), |_, e| e.map(crate::unit))?;

        Ok(None)
    }

    fn visit_parameter_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        ty: Index,
        _name: intern::Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, ty);
        let info = self.ip.size_of_type(ty, AMD64_POINTER_TYPE_INFO);
        let ir = self.ir.push(
            Inst::Parameter(ty),
            Some(triples::Data::new(info.size(), info.align())),
        );
        self.ref_lookup.insert(idx, ir);

        Ok(None)
    }

    fn visit_address_of_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        expr: Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, expr);
        let expr = self.visit_any(ast, expr)?.unwrap();
        let r = self
            .ir
            .push(Inst::GetElementPtr(ty), Some(triples::Data::new(expr, 0)));

        Ok(Some(r))
    }

    fn visit_assign_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, idx);
        let dst = self.visit_any(ast, lhs)?.unwrap();
        let src = self.visit_any(ast, rhs)?.unwrap();

        let ir = self
            .ir
            .push(Inst::Store(ty.into()), Some(triples::Data::new(src, dst)));

        Ok(Some(ir))
    }

    fn visit_explicit_cast_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        expr: Index,
        ty: Index,
    ) -> Result<Self::Value, Self::Error> {
        let lty = ast.get_type_of_node(&self.ip, &mut self.type_cache, expr);
        let rty = ast.get_type_of_node(&self.ip, &mut self.type_cache, ty);

        let lhs = self.visit_any(ast, expr)?.unwrap();

        // TODO: if bitwidth is the same, then this is a noop?
        let cast = self
            .ir
            .push(Inst::ExplicitCast(lty, rty), Some(triples::Data::lhs(lhs)));

        Ok(Some(cast))
    }

    fn visit_if_expr_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        cond: Index,
        body: Index,
    ) -> Result<Self::Value, Self::Error> {
        let cond = self.visit_any(ast, cond)?.unwrap();
        let br = self.ir.push(Inst::Branch(cond), None);

        let lbody = self
            .ir
            .push(Inst::Label, Some(intern::Index::EMPTY_STRING.into()));
        _ = self.visit_any(ast, body);
        let jmp = self.ir.push(Inst::Jump, None);
        let skip = self
            .ir
            .push(Inst::Label, Some(intern::Index::EMPTY_STRING.into()));

        self.ir.data[br as usize] = Some(triples::Data::new(lbody, skip));
        self.ir.data[jmp as usize] = Some(triples::Data::lhs(skip));

        Ok(Some(br))
    }

    fn visit_if_else_expr_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        cond: Index,
        a: Index,
        b: Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, idx);

        let cond = self.visit_any(ast, cond)?.unwrap();
        let br = self.ir.push(Inst::Branch(cond), None);

        let la = self
            .ir
            .push(Inst::Label, Some(intern::Index::EMPTY_STRING.into()));
        let a = self.visit_any(ast, a)?.unwrap();
        let jmpa = self.ir.push(Inst::Jump, None);

        let lb = self
            .ir
            .push(Inst::Label, Some(intern::Index::EMPTY_STRING.into()));
        let b = self.visit_any(ast, b)?.unwrap();
        let jmpb = self.ir.push(Inst::Jump, None);

        let lphi = self
            .ir
            .push(Inst::Label, Some(intern::Index::EMPTY_STRING.into()));

        let phi = if ty == intern::Index::VOID {
            br
        } else {
            self.ir.push(Inst::Phi2(ty), Some(triples::Data::new(a, b)))
        };

        self.ir.data[br as usize] = Some(triples::Data::new(la, lb));
        self.ir.data[jmpa as usize] = Some(triples::Data::lhs(lphi));
        self.ir.data[jmpb as usize] = Some(triples::Data::lhs(lphi));

        Ok(Some(phi))
    }

    fn visit_deref_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        expr: Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, idx);
        let expr = self.visit_any(ast, expr)?.unwrap();
        let not = self.ir.push(Inst::Load(ty), Some(triples::Data::lhs(expr)));

        Ok(Some(not))
    }

    fn visit_not_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        expr: Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, expr);
        let expr = self.visit_any(ast, expr)?.unwrap();
        let not = self.ir.push(Inst::Not(ty), Some(triples::Data::lhs(expr)));

        Ok(Some(not))
    }

    fn visit_negate_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        expr: Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, expr);
        let expr = self.visit_any(ast, expr)?.unwrap();
        let not = self
            .ir
            .push(Inst::Negate(ty), Some(triples::Data::lhs(expr)));

        Ok(Some(not))
    }

    fn visit_or_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, idx);

        // TODO: make this an error instead of a panic
        let a = self.visit_any(ast, lhs)?.unwrap();
        let not_a = self.ir.push(Inst::Not(ty), Some(triples::Data::lhs(a)));

        // branch taken if a
        let br = self.ir.push(Inst::Branch(not_a), None);
        let lb = self
            .ir
            .push(Inst::Label, Some(intern::Index::EMPTY_STRING.into()));

        let b = self.visit_any(ast, rhs)?.unwrap();
        let jmp = self.ir.push(Inst::Jump, None);

        let la = self
            .ir
            .push(Inst::Label, Some(intern::Index::EMPTY_STRING.into()));

        self.ir.data[br as usize] = Some(triples::Data::new(lb, la));
        self.ir.data[jmp as usize] = Some(triples::Data::lhs(la));

        let or = self.ir.push(
            Inst::Phi2(intern::Index::BOOL),
            Some(triples::Data::new(a, b)),
        );

        Ok(Some(or))
    }

    fn visit_and_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        // TODO: make this an error instead of a panic
        let a = self.visit_any(ast, lhs)?.unwrap();

        // branch taken if a
        let br = self.ir.push(Inst::Branch(a), None);

        let lb = self
            .ir
            .push(Inst::Label, Some(intern::Index::EMPTY_STRING.into()));
        let b = self.visit_any(ast, rhs)?.unwrap();

        let skip = self
            .ir
            .push(Inst::Label, Some(intern::Index::EMPTY_STRING.into()));

        self.ir.data[br as usize] = Some(triples::Data::new(lb, skip));
        let and = self.ir.push(
            Inst::Phi2(intern::Index::BOOL),
            Some(triples::Data::new(a, b)),
        );

        Ok(Some(and))
    }

    fn visit_add_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_sub_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_div_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_rem_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_mul_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_bit_and_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_bit_or_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_bit_x_or_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_eq_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_n_eq_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_lt_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_gt_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_le_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_ge_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_shl_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
    fn visit_shr_impl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        self.visit_binop_expr(ast, idx, lhs, rhs)
    }
}

impl IrFunctionBuilder<'_> {
    fn visit_var_decl_common<'a>(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
    ) -> Result<Option<u32>, Error> {
        let (tag, data) = ast.get_node_tag_and_data(idx);
        let (a, b) = data.as_extra_range();

        let range = &ast.extra[a..b];
        let _name = range.get(0).unwrap();

        let (expr, typed_index) = match tag {
            Tag::VarDecl | Tag::MutVarDecl => {
                let typed_index = range
                    .get(1)
                    .cloned()
                    .map(Index::from_u32)
                    .flatten()
                    .unwrap();
                (None, typed_index)
            }
            Tag::VarDeclAssignment | Tag::MutVarDeclAssignment => {
                let expr = range
                    .get(1)
                    .cloned()
                    .map(Index::from_u32)
                    .flatten()
                    .unwrap();
                let typed_index = range
                    .get(2)
                    .cloned()
                    .map(Index::from_u32)
                    .flatten()
                    .unwrap_or(expr);
                (Some(expr), typed_index)
            }
            _ => unreachable!(),
        };

        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, typed_index);
        let info = self.ip.size_of_type(ty, AMD64_POINTER_TYPE_INFO);

        let alloca = self.ir.push(
            Inst::Alloca,
            Some(triples::Data::new(info.size(), info.align())),
        );

        if let Some(idx) = expr {
            let expr = self.visit_any(ast, idx)?.unwrap();
            _ = self
                .ir
                .push(Inst::Store(ty), Some(triples::Data::new(expr, alloca)));
        }

        self.ref_lookup.insert(idx, alloca);

        Ok(Some(alloca))
    }

    /// handles all binary operations that don't short circuit and visit
    /// both lhs and rhs.
    fn visit_binop_expr<'a>(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Option<u32>, Error> {
        use triples::Data;
        let tag = ast.tags[idx];

        let ty = ast.get_type_of_node(&self.ip, &mut self.type_cache, idx);

        // TODO: make this an error instead of a panic
        let lhs = self.visit_any(ast, lhs)?.unwrap();
        let rhs = self.visit_any(ast, rhs)?.unwrap();

        let (inst, data) = match tag {
            Tag::BitOr => (Inst::BitOr(ty), Data::new(lhs, rhs)),
            Tag::BitXOr => (Inst::BitOr(ty), Data::new(lhs, rhs)),
            Tag::BitAnd => (Inst::BitOr(ty), Data::new(lhs, rhs)),

            Tag::Eq => (Inst::Eq(ty), Data::new(lhs, rhs)),
            Tag::NEq => (Inst::Neq(ty), Data::new(lhs, rhs)),
            Tag::Lt => (Inst::Lt(ty), Data::new(lhs, rhs)),
            Tag::Gt => (Inst::Gt(ty), Data::new(lhs, rhs)),
            Tag::Le => (Inst::Le(ty), Data::new(lhs, rhs)),
            Tag::Ge => (Inst::Ge(ty), Data::new(lhs, rhs)),

            Tag::Shl => (Inst::ShiftLeft(ty), Data::new(lhs, rhs)),
            Tag::Shr => (Inst::ShiftRight(ty), Data::new(lhs, rhs)),
            Tag::Add => (Inst::Add(ty), Data::new(lhs, rhs)),
            Tag::Sub => (Inst::Sub(ty), Data::new(lhs, rhs)),
            Tag::Div => (Inst::Div(ty), Data::new(lhs, rhs)),
            Tag::Rem => (Inst::Rem(ty), Data::new(lhs, rhs)),
            Tag::Mul => (Inst::Mul(ty), Data::new(lhs, rhs)),
            _ => panic!("not a binop"),
        };

        Ok(Some(self.ir.push(inst, Some(data))))
    }
}

impl<'a> AstVisitorTrait<&'a mut super::Ast> for IrBuilder {
    type Value = ();
    type Error = Error;
    const UNIMPL: Self::Error = Error::Unimplemented;

    fn visit_function_decl(
        &mut self,
        ast: &'a mut super::Ast,
        idx: super::Index,
    ) -> Result<(), Self::Error> {
        let data = ast.expect_node_data_for_tag(idx, Tag::FunctionDecl);
        let (_proto, _block) = data.as_two_indices();

        // visit proto

        // visit block

        todo!()
    }
}

pub struct PlacenessSolver;

impl<'a> AstVisitorTrait<&'a mut Ast> for PlacenessSolver {
    type Error = ();

    type Value = PlaceOrValue;

    const UNIMPL: Self::Error = ();

    fn visit_subscript_expr_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        lhs: Index,
        index: Index,
    ) -> Result<Self::Value, Self::Error> {
        let lhs = self.visit_any(ast, lhs)?;
        let index = self.visit_any(ast, index)?;

        let lhs = ast.convert_to_value_expr(lhs);
        let index = ast.convert_to_value_expr(index);

        ast.datas[idx] = Data::two_indices(lhs, index);

        Ok(PlaceOrValue::Place(idx))
    }

    fn visit_deref_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        expr: Index,
    ) -> Result<Self::Value, Self::Error> {
        let expr = self.visit_any(ast, expr)?;
        let expr = ast.convert_to_value_expr(expr);

        ast.datas[idx] = Data::index(expr);

        Ok(PlaceOrValue::Place(idx))
    }

    fn visit_parameter_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        ty: Index,
        name: intern::Index,
    ) -> Result<Self::Value, Self::Error> {
        Ok(PlaceOrValue::Value(idx))
    }

    fn visit_var_decl_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        name: intern::Index,
        ty: Index,
    ) -> Result<Self::Value, Self::Error> {
        Ok(PlaceOrValue::Place(idx))
    }

    fn visit_global_decl_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        name: intern::Index,
        expr: Index,
        ty: Index,
    ) -> Result<Self::Value, Self::Error> {
        Ok(PlaceOrValue::Place(idx))
    }

    fn visit_address_of_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        expr: Index,
    ) -> Result<Self::Value, Self::Error> {
        Ok(PlaceOrValue::Value(idx))
    }

    fn visit_decl_ref_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        decl: Index,
    ) -> Result<Self::Value, Self::Error> {
        let decl = self.visit_any(ast, decl)?;

        // copy the placeness of the decl referenced.
        // parameters are values, var decls are places
        Ok(decl.with_index(idx))
    }

    fn visit_function_decl_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        proto: Index,
        body: Index,
    ) -> Result<Self::Value, Self::Error> {
        let _body = self.visit_block_maybe_trailing_as_value(ast, body)?;

        Ok(PlaceOrValue::Value(idx))
    }

    fn visit_block_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        statements: Vec<Index>,
        expr: Option<Index>,
    ) -> Result<Self::Value, Self::Error> {
        for child in statements {
            self.visit_any(ast, child)?;
        }

        if let Some(expr) = expr {
            // copy placeness of trailing expression
            Ok(self.visit_any(ast, expr)?.with_index(idx))
        } else {
            // this block returns nothing, so it can safely be a value
            Ok(PlaceOrValue::Value(idx))
        }
    }

    fn visit_return_expr_stmt_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        expr: Index,
    ) -> Result<Self::Value, Self::Error> {
        let expr = self.visit_any(ast, expr)?;
        let expr = ast.convert_to_value_expr(expr);

        ast.datas[idx] = Data::index(expr);

        Ok(PlaceOrValue::Value(idx))
    }

    fn visit_assign_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        lhs: Index,
        rhs: Index,
    ) -> Result<Self::Value, Self::Error> {
        let (dst, src) = (lhs, rhs);

        let dst = self.visit_any(ast, dst)?;
        let src = self.visit_any(ast, src)?;

        let dst = ast.convert_to_place_expr(dst);
        let src = ast.convert_to_value_expr(src);

        ast.datas[idx] = Data::two_indices(dst, src);

        Ok(PlaceOrValue::Value(idx))
    }

    fn visit_if_expr_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        cond: Index,
        body: Index,
    ) -> Result<Self::Value, Self::Error> {
        let cond = self.visit_any(ast, cond)?;
        let body = self.visit_any(ast, body)?;

        let cond = ast.convert_to_value_expr(cond);

        ast.datas[idx] = Data::two_indices(cond, body.into_index());

        Ok(body.with_index(idx))
    }

    fn visit_if_else_expr_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        cond: Index,
        a: Index,
        b: Index,
    ) -> Result<Self::Value, Self::Error> {
        let (_, data) = ast.get_node_tag_and_data(idx);
        let (_, extra) = data.as_index_and_extra_offset();

        let cond = self.visit_any(ast, cond)?;
        let cond = ast.convert_to_value_expr(cond);

        let a = self.visit_any(ast, a)?;
        let b = self.visit_any(ast, b)?;

        // if placeness differs between branches, convert both to values.
        let bodies = if !a.eq_discriminant(&b) {
            [ast.convert_to_value_expr(a), ast.convert_to_value_expr(b)]
        } else {
            [a.into_index(), b.into_index()]
        }
        .map(|a| a.as_u32());

        ast.datas[idx] = Data::index_and_extra_offset(cond, extra as u32);
        ast.extra[extra..][..2].copy_from_slice(&bodies);

        Ok(a.with_index(idx))
    }

    fn visit_call_expr_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        func: Index,
        argument_list: Index,
    ) -> Result<Self::Value, Self::Error> {
        let func = self.visit_any(ast, func)?;
        let func = ast.convert_to_value_expr(func);
        let arguments = self.visit_argument_list(ast, argument_list)?;

        ast.datas[idx] = Data::two_indices(func, arguments.into_index());

        Ok(PlaceOrValue::Value(idx))
    }

    fn visit_argument_list_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        arguments: Vec<Index>,
    ) -> Result<Self::Value, Self::Error> {
        for child in arguments {
            self.visit_argument(ast, child)?;
        }

        Ok(PlaceOrValue::Value(idx))
    }

    fn visit_argument_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        expr: Index,
        name: Option<intern::Index>,
    ) -> Result<Self::Value, Self::Error> {
        self.value_index(ast, idx)
    }

    fn visit_explicit_cast_impl(
        &mut self,
        ast: &'a mut Ast,
        idx: Index,
        lhs: Index,
        ty: Index,
    ) -> Result<Self::Value, Self::Error> {
        let lhs = self.visit_any(ast, lhs)?;
        let lhs = ast.convert_to_value_expr(lhs);

        ast.datas[idx] = Data::two_indices(lhs, ty);

        Ok(PlaceOrValue::Value(idx))
    }

    fn visit_any(&mut self, ast: &'a mut Ast, idx: Index) -> Result<Self::Value, Self::Error> {
        let tag = ast.tags[idx];

        let idx = match tag {
            Tag::Parameter => self.visit_parameter(ast, idx)?,
            Tag::VarDecl | Tag::MutVarDecl | Tag::VarDeclAssignment | Tag::MutVarDeclAssignment => {
                self.visit_var_decl_common(ast, idx)?
            }
            Tag::GlobalDecl => self.visit_global_decl(ast, idx)?,
            Tag::DeclRef => self.visit_decl_ref(ast, idx)?,
            Tag::Deref => self.visit_deref(ast, idx)?,
            Tag::AddressOf => self.visit_address_of(ast, idx)?,
            Tag::SubscriptExpr => self.visit_subscript_expr(ast, idx)?,
            Tag::FunctionDecl => self.visit_function_decl(ast, idx)?,
            Tag::BlockTrailingExpr | Tag::Block => self.visit_block(ast, idx)?,
            Tag::ExprStmt | Tag::Constant | Tag::ReturnStmt => PlaceOrValue::Value(idx),
            Tag::ReturnExprStmt => self.visit_return_expr_stmt(ast, idx)?,
            Tag::ArgumentList => self.visit_argument_list(ast, idx)?,
            Tag::ExplicitCast => self.visit_explicit_cast(ast, idx)?,
            Tag::CallExpr => self.visit_call_expr(ast, idx)?,
            Tag::NamedArgument | Tag::Argument => self.visit_argument(ast, idx)?,
            Tag::Not | Tag::Negate => self.value_index(ast, idx)?,
            Tag::Or
            | Tag::And
            | Tag::BitOr
            | Tag::BitXOr
            | Tag::BitAnd
            | Tag::Eq
            | Tag::NEq
            | Tag::Lt
            | Tag::Gt
            | Tag::Le
            | Tag::Ge
            | Tag::Shl
            | Tag::Shr
            | Tag::Add
            | Tag::Sub
            | Tag::Mul
            | Tag::Div
            | Tag::Rem => self.value_two_indices(ast, idx)?,
            Tag::Assign => self.visit_assign(ast, idx)?,
            Tag::IfExpr => self.visit_if_expr(ast, idx)?,
            Tag::IfElseExpr => self.visit_if_else_expr(ast, idx)?,
            Tag::File | Tag::Root => self.visit_children(ast, idx)?,

            Tag::FieldAccess => todo!(),
            Tag::PlaceToValueConversion => todo!(),
            Tag::ValueToPlaceConversion => todo!(),
            Tag::StructDecl => todo!(),
            Tag::FieldDecl => todo!(),
            Tag::TypeDeclRef => todo!(),

            Tag::DeclRefUnresolved => todo!(),
            Tag::InternedType => todo!(),
            Tag::TypeDeclRefUnresolved => todo!(),
            Tag::PointerType => todo!(),
            Tag::ArrayType => todo!(),
            Tag::Error => todo!(),
            Tag::Undefined => todo!(),
            Tag::FunctionProto => todo!(),
            Tag::FunctionProtoInterned => todo!(),
            Tag::StructDeclInterned => todo!(),
            Tag::ParameterList => todo!(),
        };
        //_ => Err(())?,

        Ok(idx)
    }
}

impl PlacenessSolver {
    fn value_two_indices<'a>(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
    ) -> Result<PlaceOrValue, ()> {
        let (a, b) = ast.datas[idx].as_two_indices();

        let a = self.visit_any(ast, a)?;
        let b = self.visit_any(ast, b)?;

        let a = ast.convert_to_value_expr(a);
        let b = ast.convert_to_value_expr(b);

        ast.datas[idx] = Data::two_indices(a, b);

        Ok(PlaceOrValue::Value(idx))
    }

    fn value_index<'a>(&mut self, ast: &'a mut super::Ast, idx: Index) -> Result<PlaceOrValue, ()> {
        let a = ast.datas[idx].as_index();

        let a = self.visit_any(ast, a)?;

        let a = ast.convert_to_value_expr(a);

        ast.datas[idx] = Data::index(a);

        Ok(PlaceOrValue::Value(idx))
    }

    /// this ensures the trailing expression of a block is a value-expression
    fn visit_block_maybe_trailing_as_value<'a>(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
    ) -> Result<PlaceOrValue, ()> {
        match ast.tags[idx] {
            Tag::BlockTrailingExpr => {
                let (a, b) = ast.datas[idx].as_extra_range();
                let children = unsafe { Index::from_slice_unchecked(&ast.extra[a..b]).to_vec() };
                let (expr, statements) = children.split_last().unwrap();

                for statement in statements {
                    self.visit_any(ast, *statement)?;
                }

                let expr = self.visit_any(ast, *expr)?;
                ast.extra[b - 1] = ast.convert_to_value_expr(expr).into_u32();
            }
            Tag::Block => {
                self.visit_block(ast, idx)?;
            }
            _ => unreachable!(),
        };

        Ok(PlaceOrValue::Value(idx))
    }

    fn visit_var_decl_common<'a>(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
    ) -> Result<PlaceOrValue, ()> {
        self.visit_children(ast, idx)?;
        Ok(PlaceOrValue::Place(idx))
    }

    fn visit_children<'a>(
        &mut self,
        ast: &'a mut super::Ast,
        idx: Index,
    ) -> Result<PlaceOrValue, ()> {
        for child in ast.get_node_children(idx) {
            self.visit_any(ast, child)?;
        }

        Ok(PlaceOrValue::Value(idx))
    }
}
